package com.sun.security.sasl.digest;

import javax.crypto.IllegalBlockSizeException;
import java.security.InvalidAlgorithmParameterException;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import java.util.Arrays;
import java.util.logging.Logger;
import java.security.spec.InvalidKeySpecException;
import java.security.InvalidKeyException;
import javax.crypto.spec.DESedeKeySpec;
import javax.crypto.spec.DESKeySpec;
import javax.crypto.SecretKeyFactory;
import javax.crypto.SecretKey;
import java.util.List;
import java.io.IOException;
import java.security.MessageDigest;
import javax.crypto.NoSuchPaddingException;
import java.security.NoSuchAlgorithmException;
import java.util.logging.Level;
import javax.crypto.Cipher;
import java.io.UnsupportedEncodingException;
import java.io.ByteArrayOutputStream;
import java.util.Random;
import javax.security.sasl.SaslException;
import java.util.Map;
import java.math.BigInteger;
import javax.security.auth.callback.CallbackHandler;
import com.sun.security.sasl.util.AbstractSaslImpl;

abstract class DigestMD5Base extends AbstractSaslImpl
{
    private static final String DI_CLASS_NAME;
    private static final String DP_CLASS_NAME;
    protected static final int MAX_CHALLENGE_LENGTH = 2048;
    protected static final int MAX_RESPONSE_LENGTH = 4096;
    protected static final int DEFAULT_MAXBUF = 65536;
    protected static final int DES3 = 0;
    protected static final int RC4 = 1;
    protected static final int DES = 2;
    protected static final int RC4_56 = 3;
    protected static final int RC4_40 = 4;
    protected static final String[] CIPHER_TOKENS;
    private static final String[] JCE_CIPHER_NAME;
    protected static final byte DES_3_STRENGTH = 4;
    protected static final byte RC4_STRENGTH = 4;
    protected static final byte DES_STRENGTH = 2;
    protected static final byte RC4_56_STRENGTH = 2;
    protected static final byte RC4_40_STRENGTH = 1;
    protected static final byte UNSET = 0;
    protected static final byte[] CIPHER_MASKS;
    private static final String SECURITY_LAYER_MARKER = ":00000000000000000000000000000000";
    protected static final byte[] EMPTY_BYTE_ARRAY;
    protected int step;
    protected CallbackHandler cbh;
    protected SecurityCtx secCtx;
    protected byte[] H_A1;
    protected byte[] nonce;
    protected String negotiatedStrength;
    protected String negotiatedCipher;
    protected String negotiatedQop;
    protected String negotiatedRealm;
    protected boolean useUTF8;
    protected String encoding;
    protected String digestUri;
    protected String authzid;
    private static final char[] pem_array;
    private static final int RAW_NONCE_SIZE = 30;
    private static final int ENCODED_NONCE_SIZE = 40;
    private static final BigInteger MASK;
    
    protected DigestMD5Base(Map<String, ?> props, String className, int step, String digestUri, CallbackHandler cbh) throws SaslException {
        super(props, className);
        this.useUTF8 = false;
        this.encoding = "8859_1";
        this.step = step;
        this.digestUri = digestUri;
        this.cbh = cbh;
    }
    
    public String getMechanismName() {
        return "DIGEST-MD5";
    }
    
    public byte[] unwrap(byte[] array, int n, int n2) throws SaslException {
        if (!this.completed) {
            throw new IllegalStateException("DIGEST-MD5 authentication not completed");
        }
        if (this.secCtx == null) {
            throw new IllegalStateException("Neither integrity nor privacy was negotiated");
        }
        return this.secCtx.unwrap(array, n, n2);
    }
    
    public byte[] wrap(byte[] array, int n, int n2) throws SaslException {
        if (!this.completed) {
            throw new IllegalStateException("DIGEST-MD5 authentication not completed");
        }
        if (this.secCtx == null) {
            throw new IllegalStateException("Neither integrity nor privacy was negotiated");
        }
        return this.secCtx.wrap(array, n, n2);
    }
    
    public void dispose() throws SaslException {
        if (this.secCtx != null) {
            this.secCtx = null;
        }
    }
    
    @Override
    public Object getNegotiatedProperty(String propName) {
        if (this.completed) {
            if (propName.equals("javax.security.sasl.strength")) {
                return this.negotiatedStrength;
            }
            if (propName.equals("javax.security.sasl.bound.server.name")) {
                return this.digestUri.substring(this.digestUri.indexOf(47) + 1);
            }
            return super.getNegotiatedProperty(propName);
        }
        else {
            throw new IllegalStateException("DIGEST-MD5 authentication not completed");
        }
    }
    
    protected static final byte[] generateNonce() {
        Random random = new Random();
        byte[] bytes = new byte[30];
        random.nextBytes(bytes);
        byte[] array = new byte[40];
        int n = 0;
        for (int i = 0; i < bytes.length; i += 3) {
            int n2 = bytes[i];
            int n3 = bytes[i + 1];
            int n4 = bytes[i + 2];
            array[n++] = (byte)DigestMD5Base.pem_array[n2 >>> 2 & 0x3F];
            array[n++] = (byte)DigestMD5Base.pem_array[(n2 << 4 & 0x30) + (n3 >>> 4 & 0xF)];
            array[n++] = (byte)DigestMD5Base.pem_array[(n3 << 2 & 0x3C) + (n4 >>> 6 & 0x3)];
            array[n++] = (byte)DigestMD5Base.pem_array[n4 & 0x3F];
        }
        return array;
    }
    
    protected static void writeQuotedStringValue(ByteArrayOutputStream byteArrayOutputStream, byte[] array) {
        for (int b : array) {
            if (needEscape((char)b)) {
                byteArrayOutputStream.write(92);
            }
            byteArrayOutputStream.write(b);
        }
    }
    
    private static boolean needEscape(String s) {
        int length = s.length();
        int i = 0;
        while (i < length) {
            if (needEscape(s.charAt(i))) {
                return true;
            }
            ++i;
        }
        return false;
    }
    
    private static boolean needEscape(char c) {
        return c == '\"' || c == '\\' || c == '\u007f' || (c >= '\0' && c <= '\u001f' && c != '\r' && c != '\t' && c != '\n');
    }
    
    protected static String quotedStringValue(String s) {
        if (needEscape(s)) {
            int length = s.length();
            char[] value = new char[length + length];
            int count = 0;
            int char1;
            for (int i = 0; i < length; value[count++] = char1, ++i) {
                char1 = s.charAt(i);
                if (needEscape((char)char1)) {
                    value[count++] = '\\';
                }
            }
            return new String(value, 0, count);
        }
        return s;
    }
    
    protected byte[] binaryToHex(byte[] array) throws UnsupportedEncodingException {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < array.length; ++i) {
            if ((array[i] & 0xFF) < 16) {
                sb.append("0" + Integer.toHexString(array[i] & 0xFF));
            }
            else {
                sb.append(Integer.toHexString(array[i] & 0xFF));
            }
        }
        return sb.toString().getBytes(this.encoding);
    }
    
    protected byte[] stringToByte_8859_1(String s) throws SaslException {
        char[] charArray = s.toCharArray();
        try {
            if (this.useUTF8) {
                int i = 0;
                while (i < charArray.length) {
                    if (charArray[i] > 'ÿ') {
                        return s.getBytes("UTF8");
                    }
                    ++i;
                }
            }
            return s.getBytes("8859_1");
        }
        catch (UnsupportedEncodingException ex) {
            throw new SaslException("cannot encode string in UTF8 or 8859-1 (Latin-1)", ex);
        }
    }
    
    protected static byte[] getPlatformCiphers() {
        byte[] array = new byte[DigestMD5Base.CIPHER_TOKENS.length];
        for (int i = 0; i < DigestMD5Base.JCE_CIPHER_NAME.length; ++i) {
            try {
                Cipher.getInstance(DigestMD5Base.JCE_CIPHER_NAME[i]);
                DigestMD5Base.logger.log(Level.FINE, "DIGEST01:Platform supports {0}", DigestMD5Base.JCE_CIPHER_NAME[i]);
                int n = i;
                array[n] |= DigestMD5Base.CIPHER_MASKS[i];
            }
            catch (NoSuchAlgorithmException ex) {}
            catch (NoSuchPaddingException ex2) {}
        }
        if (array[1] != 0) {
            int n2 = 3;
            array[n2] |= DigestMD5Base.CIPHER_MASKS[3];
            int n3 = 4;
            array[n3] |= DigestMD5Base.CIPHER_MASKS[4];
            return array;
        }
        return array;
    }
    
    protected byte[] generateResponseValue(String str, String str2, String param1, String s, String s2, char[] value, byte[] array, byte[] array2, int n, byte[] b) throws NoSuchAlgorithmException, UnsupportedEncodingException, IOException {
        MessageDigest instance = MessageDigest.getInstance("MD5");
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
        byteArrayOutputStream.write((str + ":" + str2).getBytes(this.encoding));
        if (param1.equals("auth-conf") || param1.equals("auth-int")) {
            DigestMD5Base.logger.log(Level.FINE, "DIGEST04:QOP: {0}", param1);
            byteArrayOutputStream.write(":00000000000000000000000000000000".getBytes(this.encoding));
        }
        if (DigestMD5Base.logger.isLoggable(Level.FINE)) {
            DigestMD5Base.logger.log(Level.FINE, "DIGEST05:A2: {0}", byteArrayOutputStream.toString());
        }
        instance.update(byteArrayOutputStream.toByteArray());
        byte[] binaryToHex = this.binaryToHex(instance.digest());
        if (DigestMD5Base.logger.isLoggable(Level.FINE)) {
            DigestMD5Base.logger.log(Level.FINE, "DIGEST06:HEX(H(A2)): {0}", new String(binaryToHex));
        }
        ByteArrayOutputStream byteArrayOutputStream2 = new ByteArrayOutputStream();
        byteArrayOutputStream2.write(this.stringToByte_8859_1(s));
        byteArrayOutputStream2.write(58);
        byteArrayOutputStream2.write(this.stringToByte_8859_1(s2));
        byteArrayOutputStream2.write(58);
        byteArrayOutputStream2.write(this.stringToByte_8859_1(new String(value)));
        instance.update(byteArrayOutputStream2.toByteArray());
        byte[] digest = instance.digest();
        if (DigestMD5Base.logger.isLoggable(Level.FINE)) {
            DigestMD5Base.logger.log(Level.FINE, "DIGEST07:H({0}) = {1}", new Object[] { byteArrayOutputStream2.toString(), new String(this.binaryToHex(digest)) });
        }
        ByteArrayOutputStream byteArrayOutputStream3 = new ByteArrayOutputStream();
        byteArrayOutputStream3.write(digest);
        byteArrayOutputStream3.write(58);
        byteArrayOutputStream3.write(array);
        byteArrayOutputStream3.write(58);
        byteArrayOutputStream3.write(array2);
        if (b != null) {
            byteArrayOutputStream3.write(58);
            byteArrayOutputStream3.write(b);
        }
        instance.update(byteArrayOutputStream3.toByteArray());
        byte[] digest2 = instance.digest();
        this.H_A1 = digest2;
        byte[] binaryToHex2 = this.binaryToHex(digest2);
        if (DigestMD5Base.logger.isLoggable(Level.FINE)) {
            DigestMD5Base.logger.log(Level.FINE, "DIGEST08:H(A1) = {0}", new String(binaryToHex2));
        }
        ByteArrayOutputStream byteArrayOutputStream4 = new ByteArrayOutputStream();
        byteArrayOutputStream4.write(binaryToHex2);
        byteArrayOutputStream4.write(58);
        byteArrayOutputStream4.write(array);
        byteArrayOutputStream4.write(58);
        byteArrayOutputStream4.write(nonceCountToHex(n).getBytes(this.encoding));
        byteArrayOutputStream4.write(58);
        byteArrayOutputStream4.write(array2);
        byteArrayOutputStream4.write(58);
        byteArrayOutputStream4.write(param1.getBytes(this.encoding));
        byteArrayOutputStream4.write(58);
        byteArrayOutputStream4.write(binaryToHex);
        if (DigestMD5Base.logger.isLoggable(Level.FINE)) {
            DigestMD5Base.logger.log(Level.FINE, "DIGEST09:KD: {0}", byteArrayOutputStream4.toString());
        }
        instance.update(byteArrayOutputStream4.toByteArray());
        byte[] binaryToHex3 = this.binaryToHex(instance.digest());
        if (DigestMD5Base.logger.isLoggable(Level.FINE)) {
            DigestMD5Base.logger.log(Level.FINE, "DIGEST10:response-value: {0}", new String(binaryToHex3));
            return binaryToHex3;
        }
        return binaryToHex3;
    }
    
    protected static String nonceCountToHex(int i) {
        String hexString = Integer.toHexString(i);
        StringBuffer sb = new StringBuffer();
        if (hexString.length() < 8) {
            for (int j = 0; j < 8 - hexString.length(); ++j) {
                sb.append("0");
            }
        }
        return sb.toString() + hexString;
    }
    
    protected static byte[][] parseDirectives(byte[] array, String[] array2, List<byte[]> list, int n) throws SaslException {
        byte[][] array3 = new byte[array2.length][];
        ByteArrayOutputStream obj = new ByteArrayOutputStream(10);
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream(10);
        int n2 = 1;
        int n3 = 0;
        int n4 = 0;
        int i = skipLws(array, 0);
        while (i < array.length) {
            int b = array[i];
            if (n2 != 0) {
                if (b == 44) {
                    if (obj.size() != 0) {
                        throw new SaslException("Directive key contains a ',':" + obj);
                    }
                    i = skipLws(array, i + 1);
                }
                else if (b == 61) {
                    if (obj.size() == 0) {
                        throw new SaslException("Empty directive key");
                    }
                    n2 = 0;
                    i = skipLws(array, i + 1);
                    if (i < array.length) {
                        if (array[i] != 34) {
                            continue;
                        }
                        n3 = 1;
                        ++i;
                    }
                    else {
                        throw new SaslException("Valueless directive found: " + obj.toString());
                    }
                }
                else if (isLws((byte)b)) {
                    i = skipLws(array, i + 1);
                    if (i < array.length) {
                        if (array[i] == 61) {
                            continue;
                        }
                        throw new SaslException("'=' expected after key: " + obj.toString());
                    }
                    else {
                        throw new SaslException("'=' expected after key: " + obj.toString());
                    }
                }
                else {
                    obj.write(b);
                    ++i;
                }
            }
            else if (n3 != 0) {
                if (b == 92) {
                    if (++i < array.length) {
                        byteArrayOutputStream.write((int)array[i]);
                        ++i;
                    }
                    else {
                        throw new SaslException("Unmatched quote found for directive: " + obj.toString() + " with value: " + byteArrayOutputStream.toString());
                    }
                }
                else if (b == 34) {
                    ++i;
                    n3 = 0;
                    n4 = 1;
                }
                else {
                    byteArrayOutputStream.write(b);
                    ++i;
                }
            }
            else if (isLws((byte)b) || b == 44) {
                extractDirective(obj.toString(), byteArrayOutputStream.toByteArray(), array2, array3, list, n);
                obj.reset();
                byteArrayOutputStream.reset();
                n2 = 1;
                n4 = (n3 = 0);
                i = skipLws(array, i + 1);
            }
            else {
                if (n4 != 0) {
                    throw new SaslException("Expecting comma or linear whitespace after quoted string: \"" + byteArrayOutputStream.toString() + "\"");
                }
                byteArrayOutputStream.write(b);
                ++i;
            }
        }
        if (n3 != 0) {
            throw new SaslException("Unmatched quote found for directive: " + obj.toString() + " with value: " + byteArrayOutputStream.toString());
        }
        if (obj.size() > 0) {
            extractDirective(obj.toString(), byteArrayOutputStream.toByteArray(), array2, array3, list, n);
            return array3;
        }
        return array3;
    }
    
    private static boolean isLws(byte b) {
        switch (b) {
            case 9:
            case 10:
            case 13:
            case 32: {
                return true;
            }
            default: {
                return false;
            }
        }
    }
    
    private static int skipLws(byte[] array, int i) {
        while (i < array.length) {
            if (!isLws(array[i])) {
                return i;
            }
            ++i;
        }
        return i;
    }
    
    private static void extractDirective(String str, byte[] bytes, String[] array, byte[][] array2, List<byte[]> list, int n) throws SaslException {
        int i = 0;
        while (i < array.length) {
            if (str.equalsIgnoreCase(array[i])) {
                if (array2[i] == null) {
                    array2[i] = bytes;
                    if (DigestMD5Base.logger.isLoggable(Level.FINE)) {
                        DigestMD5Base.logger.log(Level.FINE, "DIGEST11:Directive {0} = {1}", new Object[] { array[i], new String(array2[i]) });
                        break;
                    }
                    break;
                }
                else if (list != null && i == n) {
                    if (list.isEmpty()) {
                        list.add(array2[i]);
                    }
                    list.add(bytes);
                    break;
                }
                else {
                    throw new SaslException("DIGEST-MD5: peer sent more than one " + str + " directive: " + new String(bytes));
                }
            }
            else {
                ++i;
            }
        }
    }
    
    private static void setParityBit(byte[] array) {
        for (int i = 0; i < array.length; ++i) {
            int j = array[i] & 0xFE;
            array[i] = (byte)(j | ((Integer.bitCount(j) & 0x1) ^ 0x1));
        }
    }
    
    private static byte[] addDesParity(byte[] array, int n, int i) {
        if (i != 7) {
            throw new IllegalArgumentException("Invalid length of DES Key Value:" + i);
        }
        byte[] val = new byte[7];
        System.arraycopy(array, n, val, 0, i);
        byte[] parityBit = new byte[8];
        BigInteger shiftRight = new BigInteger(val);
        for (int j = parityBit.length - 1; j >= 0; --j) {
            parityBit[j] = shiftRight.and(DigestMD5Base.MASK).toByteArray()[0];
            int n2 = j;
            parityBit[n2] <<= 1;
            shiftRight = shiftRight.shiftRight(7);
        }
        setParityBit(parityBit);
        return parityBit;
    }
    
    private static SecretKey makeDesKeys(byte[] array, String s) throws NoSuchAlgorithmException, InvalidKeyException, InvalidKeySpecException {
        byte[] addDesParity = addDesParity(array, 0, 7);
        SecretKeyFactory instance = SecretKeyFactory.getInstance(s);
        int n = -1;
        switch (s.hashCode()) {
            case 99346: {
                if (s.equals("des")) {
                    n = 0;
                    break;
                }
                break;
            }
            case -1335250348: {
                if (s.equals("desede")) {
                    n = 1;
                    break;
                }
                break;
            }
        }
        DESKeySpec keySpec;
        switch (n) {
            case 0: {
                keySpec = new DESKeySpec(addDesParity, 0);
                if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                    AbstractSaslImpl.traceOutput(DigestMD5Base.DP_CLASS_NAME, "makeDesKeys", "DIGEST42:DES key input: ", array);
                    AbstractSaslImpl.traceOutput(DigestMD5Base.DP_CLASS_NAME, "makeDesKeys", "DIGEST43:DES key parity-adjusted: ", addDesParity);
                    AbstractSaslImpl.traceOutput(DigestMD5Base.DP_CLASS_NAME, "makeDesKeys", "DIGEST44:DES key material: ", ((DESKeySpec)keySpec).getKey());
                    DigestMD5Base.logger.log(Level.FINEST, "DIGEST45: is parity-adjusted? {0}", Boolean.valueOf(DESKeySpec.isParityAdjusted(addDesParity, 0)));
                    break;
                }
                break;
            }
            case 1: {
                byte[] addDesParity2 = addDesParity(array, 7, 7);
                byte[] key = new byte[addDesParity.length * 2 + addDesParity2.length];
                System.arraycopy(addDesParity, 0, key, 0, addDesParity.length);
                System.arraycopy(addDesParity2, 0, key, addDesParity.length, addDesParity2.length);
                System.arraycopy(addDesParity, 0, key, addDesParity.length + addDesParity2.length, addDesParity.length);
                keySpec = new DESedeKeySpec(key, 0);
                if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                    AbstractSaslImpl.traceOutput(DigestMD5Base.DP_CLASS_NAME, "makeDesKeys", "DIGEST46:3DES key input: ", array);
                    AbstractSaslImpl.traceOutput(DigestMD5Base.DP_CLASS_NAME, "makeDesKeys", "DIGEST47:3DES key ede: ", key);
                    AbstractSaslImpl.traceOutput(DigestMD5Base.DP_CLASS_NAME, "makeDesKeys", "DIGEST48:3DES key material: ", ((DESedeKeySpec)keySpec).getKey());
                    DigestMD5Base.logger.log(Level.FINEST, "DIGEST49: is parity-adjusted? ", Boolean.valueOf(DESedeKeySpec.isParityAdjusted(key, 0)));
                    break;
                }
                break;
            }
            default: {
                throw new IllegalArgumentException("Invalid DES strength:" + s);
            }
        }
        return instance.generateSecret(keySpec);
    }
    
    static void access$000(int num, byte[] buf, int start, int count) {
        AbstractSaslImpl.intToNetworkByteOrder(num, buf, start, count);
    }
    
    static void access$300(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$400(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$600(String srcClass, String srcMethod, String traceTag, byte[] output, int offset, int len) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output, offset, len);
    }
    
    static void access$700(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$800(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$1000(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$1200(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$1300(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$1400(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$1500(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$1600(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static int access$1800(byte[] buf, int start, int count) {
        return AbstractSaslImpl.networkByteOrderToInt(buf, start, count);
    }
    
    static int access$1900(byte[] buf, int start, int count) {
        return AbstractSaslImpl.networkByteOrderToInt(buf, start, count);
    }
    
    static int access$2000(byte[] buf, int start, int count) {
        return AbstractSaslImpl.networkByteOrderToInt(buf, start, count);
    }
    
    static void access$2100(int num, byte[] buf, int start, int count) {
        AbstractSaslImpl.intToNetworkByteOrder(num, buf, start, count);
    }
    
    static void access$2400(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$2500(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static SecretKey access$2600(byte[] array, String s) throws NoSuchAlgorithmException, InvalidKeyException, InvalidKeySpecException {
        return makeDesKeys(array, s);
    }
    
    static void access$2800(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$2900(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$3000(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$3100(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$3300(String srcClass, String srcMethod, String traceTag, byte[] output, int offset, int len) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output, offset, len);
    }
    
    static void access$3400(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$3500(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$3700(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$3900(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$4200(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$4400(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$4500(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$4600(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$4700(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static void access$5100(String srcClass, String srcMethod, String traceTag, byte[] output) {
        AbstractSaslImpl.traceOutput(srcClass, srcMethod, traceTag, output);
    }
    
    static int access$5300(byte[] buf, int start, int count) {
        return AbstractSaslImpl.networkByteOrderToInt(buf, start, count);
    }
    
    static int access$5400(byte[] buf, int start, int count) {
        return AbstractSaslImpl.networkByteOrderToInt(buf, start, count);
    }
    
    static int access$5500(byte[] buf, int start, int count) {
        return AbstractSaslImpl.networkByteOrderToInt(buf, start, count);
    }
    
    static {
        DI_CLASS_NAME = DigestMD5Base.DigestIntegrity.class.getName();
        DP_CLASS_NAME = DigestMD5Base.DigestPrivacy.class.getName();
        CIPHER_TOKENS = new String[] { "3des", "rc4", "des", "rc4-56", "rc4-40" };
        JCE_CIPHER_NAME = new String[] { "DESede/CBC/NoPadding", "RC4", "DES/CBC/NoPadding" };
        CIPHER_MASKS = new byte[] { 4, 4, 2, 2, 1 };
        EMPTY_BYTE_ARRAY = new byte[0];
        pem_array = new char[] { 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/' };
        MASK = new BigInteger("7f", 16);
    }
    
    class DigestIntegrity implements SecurityCtx
    {
        private static final String CLIENT_INT_MAGIC = "Digest session key to client-to-server signing key magic constant";
        private static final String SVR_INT_MAGIC = "Digest session key to server-to-client signing key magic constant";
        protected byte[] myKi;
        protected byte[] peerKi;
        protected int mySeqNum;
        protected int peerSeqNum;
        protected final byte[] messageType;
        protected final byte[] sequenceNum;
        
        DigestIntegrity(boolean b) throws SaslException {
            this.mySeqNum = 0;
            this.peerSeqNum = 0;
            this.messageType = new byte[2];
            this.sequenceNum = new byte[4];
            try {
                this.generateIntegrityKeyPair(b);
            }
            catch (UnsupportedEncodingException ex) {
                throw new SaslException("DIGEST-MD5: Error encoding strings into UTF-8", ex);
            }
            catch (IOException ex2) {
                throw new SaslException("DIGEST-MD5: Error accessing buffers required to create integrity key pairs", ex2);
            }
            catch (NoSuchAlgorithmException ex3) {
                throw new SaslException("DIGEST-MD5: Unsupported digest algorithm used to create integrity key pairs", ex3);
            }
            DigestMD5Base.access$000(1, this.messageType, 0, 2);
        }
        
        private void generateIntegrityKeyPair(boolean b) throws UnsupportedEncodingException, IOException, NoSuchAlgorithmException {
            byte[] bytes = "Digest session key to client-to-server signing key magic constant".getBytes(DigestMD5Base.this.encoding);
            byte[] bytes2 = "Digest session key to server-to-client signing key magic constant".getBytes(DigestMD5Base.this.encoding);
            MessageDigest instance = MessageDigest.getInstance("MD5");
            byte[] array = new byte[DigestMD5Base.this.H_A1.length + bytes.length];
            System.arraycopy(DigestMD5Base.this.H_A1, 0, array, 0, DigestMD5Base.this.H_A1.length);
            System.arraycopy(bytes, 0, array, DigestMD5Base.this.H_A1.length, bytes.length);
            instance.update(array);
            byte[] digest = instance.digest();
            System.arraycopy(bytes2, 0, array, DigestMD5Base.this.H_A1.length, bytes2.length);
            instance.update(array);
            byte[] digest2 = instance.digest();
            if (DigestMD5Base.logger.isLoggable(Level.FINER)) {
                DigestMD5Base.access$300(DigestMD5Base.DI_CLASS_NAME, "generateIntegrityKeyPair", "DIGEST12:Kic: ", digest);
                DigestMD5Base.access$400(DigestMD5Base.DI_CLASS_NAME, "generateIntegrityKeyPair", "DIGEST13:Kis: ", digest2);
            }
            if (b) {
                this.myKi = digest;
                this.peerKi = digest2;
            }
            else {
                this.myKi = digest2;
                this.peerKi = digest;
            }
        }
        
        @Override
        public byte[] wrap(byte[] array, int n, int n2) throws SaslException {
            if (n2 == 0) {
                return DigestMD5Base.EMPTY_BYTE_ARRAY;
            }
            byte[] array2 = new byte[n2 + 10 + 2 + 4];
            System.arraycopy(array, n, array2, 0, n2);
            this.incrementSeqNum();
            byte[] hmac = this.getHMAC(this.myKi, this.sequenceNum, array, n, n2);
            if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                DigestMD5Base.access$600(DigestMD5Base.DI_CLASS_NAME, "wrap", "DIGEST14:outgoing: ", array, n, n2);
                DigestMD5Base.access$700(DigestMD5Base.DI_CLASS_NAME, "wrap", "DIGEST15:seqNum: ", this.sequenceNum);
                DigestMD5Base.access$800(DigestMD5Base.DI_CLASS_NAME, "wrap", "DIGEST16:MAC: ", hmac);
            }
            System.arraycopy(hmac, 0, array2, n2, 10);
            System.arraycopy(this.messageType, 0, array2, n2 + 10, 2);
            System.arraycopy(this.sequenceNum, 0, array2, n2 + 12, 4);
            if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                DigestMD5Base.access$1000(DigestMD5Base.DI_CLASS_NAME, "wrap", "DIGEST17:wrapped: ", array2);
                return array2;
            }
            return array2;
        }
        
        @Override
        public byte[] unwrap(byte[] array, int n, int n2) throws SaslException {
            if (n2 == 0) {
                return DigestMD5Base.EMPTY_BYTE_ARRAY;
            }
            byte[] a = new byte[10];
            byte[] array2 = new byte[n2 - 16];
            byte[] a2 = new byte[2];
            byte[] array3 = new byte[4];
            System.arraycopy(array, n, array2, 0, array2.length);
            System.arraycopy(array, n + array2.length, a, 0, 10);
            System.arraycopy(array, n + array2.length + 10, a2, 0, 2);
            System.arraycopy(array, n + array2.length + 12, array3, 0, 4);
            byte[] hmac = this.getHMAC(this.peerKi, array3, array2, 0, array2.length);
            if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                DigestMD5Base.access$1200(DigestMD5Base.DI_CLASS_NAME, "unwrap", "DIGEST18:incoming: ", array2);
                DigestMD5Base.access$1300(DigestMD5Base.DI_CLASS_NAME, "unwrap", "DIGEST19:MAC: ", a);
                DigestMD5Base.access$1400(DigestMD5Base.DI_CLASS_NAME, "unwrap", "DIGEST20:messageType: ", a2);
                DigestMD5Base.access$1500(DigestMD5Base.DI_CLASS_NAME, "unwrap", "DIGEST21:sequenceNum: ", array3);
                DigestMD5Base.access$1600(DigestMD5Base.DI_CLASS_NAME, "unwrap", "DIGEST22:expectedMAC: ", hmac);
            }
            if (!Arrays.equals(a, hmac)) {
                DigestMD5Base.logger.log(Level.INFO, "DIGEST23:Unmatched MACs");
                return DigestMD5Base.EMPTY_BYTE_ARRAY;
            }
            if (this.peerSeqNum != DigestMD5Base.access$1800(array3, 0, 4)) {
                throw new SaslException("DIGEST-MD5: Out of order sequencing of messages from server. Got: " + DigestMD5Base.access$1900(array3, 0, 4) + " Expected: " + this.peerSeqNum);
            }
            if (!Arrays.equals(this.messageType, a2)) {
                throw new SaslException("DIGEST-MD5: invalid message type: " + DigestMD5Base.access$2000(a2, 0, 2));
            }
            ++this.peerSeqNum;
            return array2;
        }
        
        protected byte[] getHMAC(byte[] key, byte[] array, byte[] array2, int n, int n2) throws SaslException {
            byte[] input = new byte[4 + n2];
            System.arraycopy(array, 0, input, 0, 4);
            System.arraycopy(array2, n, input, 4, n2);
            try {
                SecretKeySpec key2 = new SecretKeySpec(key, "HmacMD5");
                Mac instance = Mac.getInstance("HmacMD5");
                instance.init(key2);
                instance.update(input);
                byte[] doFinal = instance.doFinal();
                byte[] array3 = new byte[10];
                System.arraycopy(doFinal, 0, array3, 0, 10);
                return array3;
            }
            catch (InvalidKeyException ex) {
                throw new SaslException("DIGEST-MD5: Invalid bytes used for key of HMAC-MD5 hash.", ex);
            }
            catch (NoSuchAlgorithmException ex2) {
                throw new SaslException("DIGEST-MD5: Error creating instance of MD5 digest algorithm", ex2);
            }
        }
        
        protected void incrementSeqNum() {
            DigestMD5Base.access$2100(this.mySeqNum++, this.sequenceNum, 0, 4);
        }
    }
    
    final class DigestPrivacy extends DigestMD5Base.DigestIntegrity implements SecurityCtx
    {
        private static final String CLIENT_CONF_MAGIC = "Digest H(A1) to client-to-server sealing key magic constant";
        private static final String SVR_CONF_MAGIC = "Digest H(A1) to server-to-client sealing key magic constant";
        private Cipher encCipher;
        private Cipher decCipher;
        
        DigestPrivacy(boolean b) throws SaslException {
            super(b);
            try {
                this.generatePrivacyKeyPair(b);
            }
            catch (SaslException ex) {
                throw ex;
            }
            catch (UnsupportedEncodingException ex2) {
                throw new SaslException("DIGEST-MD5: Error encoding string value into UTF-8", ex2);
            }
            catch (IOException ex3) {
                throw new SaslException("DIGEST-MD5: Error accessing buffers required to generate cipher keys", ex3);
            }
            catch (NoSuchAlgorithmException ex4) {
                throw new SaslException("DIGEST-MD5: Error creating instance of required cipher or digest", ex4);
            }
        }
        
        private void generatePrivacyKeyPair(boolean b) throws IOException, UnsupportedEncodingException, NoSuchAlgorithmException, SaslException {
            byte[] bytes = "Digest H(A1) to client-to-server sealing key magic constant".getBytes(DigestMD5Base.this.encoding);
            byte[] bytes2 = "Digest H(A1) to server-to-client sealing key magic constant".getBytes(DigestMD5Base.this.encoding);
            MessageDigest instance = MessageDigest.getInstance("MD5");
            int n;
            if (DigestMD5Base.this.negotiatedCipher.equals(DigestMD5Base.CIPHER_TOKENS[4])) {
                n = 5;
            }
            else if (DigestMD5Base.this.negotiatedCipher.equals(DigestMD5Base.CIPHER_TOKENS[3])) {
                n = 7;
            }
            else {
                n = 16;
            }
            byte[] array = new byte[n + bytes.length];
            System.arraycopy(DigestMD5Base.this.H_A1, 0, array, 0, n);
            System.arraycopy(bytes, 0, array, n, bytes.length);
            instance.update(array);
            byte[] digest = instance.digest();
            System.arraycopy(bytes2, 0, array, n, bytes2.length);
            instance.update(array);
            byte[] digest2 = instance.digest();
            if (DigestMD5Base.logger.isLoggable(Level.FINER)) {
                DigestMD5Base.access$2400(DigestMD5Base.DP_CLASS_NAME, "generatePrivacyKeyPair", "DIGEST24:Kcc: ", digest);
                DigestMD5Base.access$2500(DigestMD5Base.DP_CLASS_NAME, "generatePrivacyKeyPair", "DIGEST25:Kcs: ", digest2);
            }
            if (b) {
            }
            else {
                digest = (digest2 = digest2);
            }
            try {
                if (DigestMD5Base.this.negotiatedCipher.indexOf(DigestMD5Base.CIPHER_TOKENS[1]) > -1) {
                    this.encCipher = Cipher.getInstance("RC4");
                    this.decCipher = Cipher.getInstance("RC4");
                    SecretKeySpec key = new SecretKeySpec(digest, "RC4");
                    SecretKeySpec key2 = new SecretKeySpec(digest2, "RC4");
                    this.encCipher.init(1, key);
                    this.decCipher.init(2, key2);
                }
                else if (DigestMD5Base.this.negotiatedCipher.equals(DigestMD5Base.CIPHER_TOKENS[2]) || DigestMD5Base.this.negotiatedCipher.equals(DigestMD5Base.CIPHER_TOKENS[0])) {
                    String s;
                    String s2;
                    if (DigestMD5Base.this.negotiatedCipher.equals(DigestMD5Base.CIPHER_TOKENS[2])) {
                        s = "DES/CBC/NoPadding";
                        s2 = "des";
                    }
                    else {
                        s = "DESede/CBC/NoPadding";
                        s2 = "desede";
                    }
                    this.encCipher = Cipher.getInstance(s);
                    this.decCipher = Cipher.getInstance(s);
                    SecretKey access$2600 = DigestMD5Base.access$2600(digest, s2);
                    SecretKey access$2601 = DigestMD5Base.access$2600(digest2, s2);
                    IvParameterSpec params = new IvParameterSpec(digest, 8, 8);
                    IvParameterSpec params2 = new IvParameterSpec(digest2, 8, 8);
                    this.encCipher.init(1, access$2600, params);
                    this.decCipher.init(2, access$2601, params2);
                    if (DigestMD5Base.logger.isLoggable(Level.FINER)) {
                        DigestMD5Base.access$2800(DigestMD5Base.DP_CLASS_NAME, "generatePrivacyKeyPair", "DIGEST26:" + DigestMD5Base.this.negotiatedCipher + " IVcc: ", params.getIV());
                        DigestMD5Base.access$2900(DigestMD5Base.DP_CLASS_NAME, "generatePrivacyKeyPair", "DIGEST27:" + DigestMD5Base.this.negotiatedCipher + " IVcs: ", params2.getIV());
                        DigestMD5Base.access$3000(DigestMD5Base.DP_CLASS_NAME, "generatePrivacyKeyPair", "DIGEST28:" + DigestMD5Base.this.negotiatedCipher + " encryption key: ", access$2600.getEncoded());
                        DigestMD5Base.access$3100(DigestMD5Base.DP_CLASS_NAME, "generatePrivacyKeyPair", "DIGEST29:" + DigestMD5Base.this.negotiatedCipher + " decryption key: ", access$2601.getEncoded());
                    }
                }
            }
            catch (InvalidKeySpecException ex) {
                throw new SaslException("DIGEST-MD5: Unsupported key specification used.", ex);
            }
            catch (InvalidAlgorithmParameterException ex2) {
                throw new SaslException("DIGEST-MD5: Invalid cipher algorithem parameter used to create cipher instance", ex2);
            }
            catch (NoSuchPaddingException ex3) {
                throw new SaslException("DIGEST-MD5: Unsupported padding used for chosen cipher", ex3);
            }
            catch (InvalidKeyException ex4) {
                throw new SaslException("DIGEST-MD5: Invalid data used to initialize keys", ex4);
            }
        }
        
        @Override
        public byte[] wrap(byte[] array, int n, int n2) throws SaslException {
            if (n2 == 0) {
                return DigestMD5Base.EMPTY_BYTE_ARRAY;
            }
            this.incrementSeqNum();
            byte[] hmac = this.getHMAC(this.myKi, this.sequenceNum, array, n, n2);
            if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                DigestMD5Base.access$3300(DigestMD5Base.DP_CLASS_NAME, "wrap", "DIGEST30:Outgoing: ", array, n, n2);
                DigestMD5Base.access$3400(DigestMD5Base.DP_CLASS_NAME, "wrap", "seqNum: ", this.sequenceNum);
                DigestMD5Base.access$3500(DigestMD5Base.DP_CLASS_NAME, "wrap", "MAC: ", hmac);
            }
            int blockSize = this.encCipher.getBlockSize();
            byte[] empty_BYTE_ARRAY;
            if (blockSize > 1) {
                int n3 = blockSize - (n2 + 10) % blockSize;
                empty_BYTE_ARRAY = new byte[n3];
                for (int i = 0; i < n3; ++i) {
                    empty_BYTE_ARRAY[i] = (byte)n3;
                }
            }
            else {
                empty_BYTE_ARRAY = DigestMD5Base.EMPTY_BYTE_ARRAY;
            }
            byte[] input = new byte[n2 + empty_BYTE_ARRAY.length + 10];
            System.arraycopy(array, n, input, 0, n2);
            System.arraycopy(empty_BYTE_ARRAY, 0, input, n2, empty_BYTE_ARRAY.length);
            System.arraycopy(hmac, 0, input, n2 + empty_BYTE_ARRAY.length, 10);
            if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                DigestMD5Base.access$3700(DigestMD5Base.DP_CLASS_NAME, "wrap", "DIGEST31:{msg, pad, KicMAC}: ", input);
            }
            byte[] update;
            try {
                update = this.encCipher.update(input);
                if (update == null) {
                    throw new IllegalBlockSizeException("" + input.length);
                }
            }
            catch (IllegalBlockSizeException ex) {
                throw new SaslException("DIGEST-MD5: Invalid block size for cipher", ex);
            }
            byte[] array2 = new byte[update.length + 2 + 4];
            System.arraycopy(update, 0, array2, 0, update.length);
            System.arraycopy(this.messageType, 0, array2, update.length, 2);
            System.arraycopy(this.sequenceNum, 0, array2, update.length + 2, 4);
            if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                DigestMD5Base.access$3900(DigestMD5Base.DP_CLASS_NAME, "wrap", "DIGEST32:Wrapped: ", array2);
                return array2;
            }
            return array2;
        }
        
        @Override
        public byte[] unwrap(byte[] array, int n, int n2) throws SaslException {
            if (n2 == 0) {
                return DigestMD5Base.EMPTY_BYTE_ARRAY;
            }
            byte[] input = new byte[n2 - 6];
            byte[] a2 = new byte[2];
            byte[] array2 = new byte[4];
            System.arraycopy(array, n, input, 0, input.length);
            System.arraycopy(array, n + input.length, a2, 0, 2);
            System.arraycopy(array, n + input.length + 2, array2, 0, 4);
            if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                DigestMD5Base.logger.log(Level.FINEST, "DIGEST33:Expecting sequence num: {0}", new Integer(this.peerSeqNum));
                DigestMD5Base.access$4200(DigestMD5Base.DP_CLASS_NAME, "unwrap", "DIGEST34:incoming: ", input);
            }
            byte[] update;
            try {
                update = this.decCipher.update(input);
                if (update == null) {
                    throw new IllegalBlockSizeException("" + input.length);
                }
            }
            catch (IllegalBlockSizeException ex) {
                throw new SaslException("DIGEST-MD5: Illegal block sizes used with chosen cipher", ex);
            }
            byte[] array3 = new byte[update.length - 10];
            byte[] a3 = new byte[10];
            System.arraycopy(update, 0, array3, 0, array3.length);
            System.arraycopy(update, array3.length, a3, 0, 10);
            if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                DigestMD5Base.access$4400(DigestMD5Base.DP_CLASS_NAME, "unwrap", "DIGEST35:Unwrapped (w/padding): ", array3);
                DigestMD5Base.access$4500(DigestMD5Base.DP_CLASS_NAME, "unwrap", "DIGEST36:MAC: ", a3);
                DigestMD5Base.access$4600(DigestMD5Base.DP_CLASS_NAME, "unwrap", "DIGEST37:messageType: ", a2);
                DigestMD5Base.access$4700(DigestMD5Base.DP_CLASS_NAME, "unwrap", "DIGEST38:sequenceNum: ", array2);
            }
            int length = array3.length;
            if (this.decCipher.getBlockSize() > 1) {
                length -= array3[array3.length - 1];
                if (length < 0) {
                    if (DigestMD5Base.logger.isLoggable(Level.INFO)) {
                        DigestMD5Base.logger.log(Level.INFO, "DIGEST39:Incorrect padding: {0}", new Byte(array3[array3.length - 1]));
                    }
                    return DigestMD5Base.EMPTY_BYTE_ARRAY;
                }
            }
            byte[] hmac = this.getHMAC(this.peerKi, array2, array3, 0, length);
            if (DigestMD5Base.logger.isLoggable(Level.FINEST)) {
                DigestMD5Base.access$5100(DigestMD5Base.DP_CLASS_NAME, "unwrap", "DIGEST40:KisMAC: ", hmac);
            }
            if (!Arrays.equals(a3, hmac)) {
                DigestMD5Base.logger.log(Level.INFO, "DIGEST41:Unmatched MACs");
                return DigestMD5Base.EMPTY_BYTE_ARRAY;
            }
            if (this.peerSeqNum != DigestMD5Base.access$5300(array2, 0, 4)) {
                throw new SaslException("DIGEST-MD5: Out of order sequencing of messages from server. Got: " + DigestMD5Base.access$5400(array2, 0, 4) + " Expected: " + this.peerSeqNum);
            }
            if (!Arrays.equals(this.messageType, a2)) {
                throw new SaslException("DIGEST-MD5: invalid message type: " + DigestMD5Base.access$5500(a2, 0, 2));
            }
            ++this.peerSeqNum;
            if (length == array3.length) {
                return array3;
            }
            byte[] array4 = new byte[length];
            System.arraycopy(array3, 0, array4, 0, length);
            return array4;
        }
    }
}
